namespace CaDiCaL {

const char *version ();
const char *copyright ();
const char *authors ();
const char *affiliations ();
const char *signature ();
const char *identifier ();
const char *compiler ();
const char *date ();
const char *flags ();

} // namespace CaDiCaL
